#include "connections.h"
#include "maptool.h"
#include "curves.h"
#include <tinyxml.h>
#include <wx/glcanvas.h>

#include <algorithm>

using std::list;
using std::multimap;

string Connection::routeFile;
list<Connection*> Connection::connections;

Connection::Connection(Location* f, Location* t, float d, ConType ty)
: from(f), to(t), distance(d), modifier(1.0f), type(ty), terrainType(0) {
	if (d==0) distance=from->pos.distance(to->pos);
}

Connection* Connection::get(const Location* a, const Location* b) {
	Connection* c=0;

	if (a && b) {
		list<Connection*>::iterator it=connections.begin();
		for (; it!=connections.end(); it++) {
			if ((*it)->from==a && (*it)->to==b) break;
		}
		c = it!=connections.end() ? *it : 0;
	}
	return c;
}

Connection* Connection::pick(const Vector3d& pos) {
	Vector3d p = pos;
	Connection* nearest=0;
	float closest=0;

	list<Connection*>::iterator it = connections.begin();
	for (; it != connections.end(); it++) {
		float dist = (*it)->distanceSq(p);
		if (!nearest || dist<closest) {
			nearest=*it;
			closest=dist;
		}
	}
	return (closest<150.0f/zoom) ? nearest : 0;
}

Connection* Connection::add(Location* a, Location* b) {
	Connection* c = 0;
	if (a && b) {
		c = new Connection(a,b);
		connections.push_front(c);

		c->via.push_back(a->pos);
		c->via.push_back(b->pos);
		a->connections.insert( ConEntry(b, c) );
		b->connections.insert( ConEntry(a, c) );

		if (selectedCon) c->type = selectedCon->type;
	}
	return c;
}

void Connection::remove(Connection* c) {
	if (c) {
		multimap<Location*, Connection*>::iterator dit, tmp;
		dit = c->from->connections.find( c->to );
		while (dit != c->from->connections.end() && dit->first == c->to) {
			tmp = dit++;
			if (tmp->second == c) c->from->connections.erase(tmp);
		}

		dit = c->to->connections.find( c->from );
		while (dit != c->to->connections.end() && dit->first == c->from) {
			tmp = dit++;
			if (tmp->second == c) c->to->connections.erase(tmp);
		}
		
		list<Connection*>::iterator it = find(connections.begin(), connections.end(), c);		
		if (it!=connections.end()) {
			delete *it;
			connections.erase(it);
		}
	}
}

void Connection::getCPointsInRect(const Vector3d& min, const Vector3d& max, list<Vector3d*>& l) {
	for (unsigned i=0; i<via.size(); i++) {
		if (via[i].x>=min.x && via[i].x<=max.x && via[i].y>=min.y && via[i].y<=max.y)
			l.push_back(&via[i]);
	}
}

float Connection::distanceSq(const Vector3d& pos) {
	float d, dist = 0;
	if (via.size()==2) 
		dist = distPointLineSegSq(pos, via[0], via[1]);
	else {
		int last=(int)via.size()-1;
		dist = CRSplinePointDistSq(pos, via[0], via[0], via[1], via[2]);
		for (int i=0; i<last-2; i++) {
			d = CRSplinePointDistSq(pos, via[i], via[i+1], via[i+2], via[i+3]);
			if (d<dist) dist=d;
		}
		d = CRSplinePointDistSq(pos, via[last-2], via[last-1], via[last], via[last]);
		if (d<dist) dist=d;
	}
	return dist;
}

float Connection::length() {
	float len = 0;
	if (via.size()==2) 
		len = via[0].distance(via[1]);
	else {
		int last=(int)via.size()-1;
		len = CRSplineLength(via[0], via[0], via[1], via[2]);
		for (int i=0; i<last-2; i++)
			len += CRSplineLength(via[i], via[i+1], via[i+2], via[i+3]);
		len += CRSplineLength(via[last-2], via[last-1], via[last], via[last]);
	}
	return len;
}

void Connection::reverse() {
	std::reverse(via.begin(), via.end());
	std::swap(from, to);
}

void Connection::insertLoc(Location* l) {
	int wp=1;
	getClosestPt(l->pos, &wp);

	Connection* newCon = add(from, l);
	newCon->terrainType=terrainType;
	newCon->type=type;
	newCon->modifier=modifier;
	for (int i=1; i<wp; i++) newCon->via.insert(newCon->via.end()-1, via[i]);

	newCon = add(l, to);
	newCon->terrainType=terrainType;
	newCon->type=type;
	newCon->modifier=modifier;
	for (int i=wp; i<(int)via.size()-1; i++) newCon->via.insert(newCon->via.end()-1, via[i]);

	selectedCon = newCon;
	remove(this);
}

Vector3d Connection::getClosestPt(const Vector3d& p, int* wpt) {
	Vector3d pos = p;
	
	Vector3d closest, v;
	int wp=1;

	if (via.size()==2) {
		closest = getClosestPtLineSeg(pos, via[0], via[1]);
	}

	else {
		int last=(int)via.size()-1;
		closest = CRGetClosestPt(pos, via[0], via[0], via[1], via[2]);
		float d, dist=closest.distanceSq(pos);
		for (int i=0; i<last-2; i++) {
			v = CRGetClosestPt(pos, via[i], via[i+1], via[i+2], via[i+3]);
			d = v.distanceSq(pos);
			if (d<dist) {
				wp=i+2;
				dist=d;
				closest=v;
			}
		}
		v = CRGetClosestPt(pos, via[last-2], via[last-1], via[last], via[last]);
		d = v.distanceSq(pos);
		if (d<dist) {
			dist=d;
			wp=last;
			closest = v;
		}
	}
	if (wpt) *wpt=wp;
	return closest;
}

void Connection::insertCP(const Vector3d& p) {
	int wp=0;
	getClosestPt(p, &wp);
	via.insert( via.begin()+wp, p );
	selectedWP=wp;
}

void Connection::insertCP() {
	if (selectedWP>0 && selectedWP<(int)via.size()-1) {
		++selectedWP;
		Vector3d half = .5*(via[selectedWP-1]+via[selectedWP]);
		via.insert( via.begin() + selectedWP, half);
	}
	else {
		Vector3d half = .5*(via.front() + *(via.begin()+1));
		via.insert( via.begin()+1, half);
		selectedWP = 1;
	}
	distance = length();
}

int Connection::pickControlPoint(const Vector3d& pos) {
	int nearest=-1;
	float closest=0;

	for (int i=1; i<(int)via.size()-1; i++) {
		float dist = pos.distanceSq(via[i]);
		if (nearest<0 || dist<closest) {
			nearest=i;
			closest=dist;
		}
	}
	if (closest>60.0f/zoom) nearest=-1;
	return nearest;
}

void Connection::deleteCP(int cp) {
	if (cp>0 && cp<(int)via.size()-1) {
		via.erase( via.begin()+cp );
		if (selectedWP >= (int)selectedCon->via.size()-1) selectedWP--;
		distance = length();
	}
}

void Connection::clear() {
	list<Connection*>::iterator it=connections.begin();
	for (; it!=connections.end(); it++) delete *it;
	connections.clear();
}

void Connection::load(const string& file) {
	TiXmlDocument xml(file.c_str());
	if (xml.LoadFile()) {
		TiXmlElement* root=xml.RootElement();
		TiXmlElement* e = root->FirstChildElement("con");

		Location *from=0, *to=0;
		int id=-1;

		clear();
		routeFile = file;

		while (e) {
			from=to=0;

			if (TIXML_WRONG_TYPE != e->QueryIntAttribute("from", &id)) from = Location::get(id);
			else from = Location::get(e->Attribute("from"));
			
			if (TIXML_WRONG_TYPE != e->QueryIntAttribute("to", &id)) to = Location::get(id);
			else to = Location::get(e->Attribute("to"));
			
			Connection* con = add(from, to);
			if (con) {
				const char* ptr = e->Attribute("type");
				if (ptr) {
					string type = ptr;
					if (type=="sea") con->type = SEA;
					if (type=="land") con->type = LAND;
					if (type=="river") con->type = RIVER;
				}
			
				TiXmlElement* v=e->FirstChildElement("via");
				while (v) {
					double x=0, y=0;
					v->QueryDoubleAttribute("x", &x);
					v->QueryDoubleAttribute("y", &y);
					con->via.insert( con->via.end()-1, Vector3d(x,y,0) );

					v=v->NextSiblingElement("via");
				}

				e->QueryValueAttribute("terrainTypes", &con->terrainType);
				con->distance = con->length();

				e->QueryFloatAttribute("dist", &con->distance);
				e->QueryFloatAttribute("modifier", &con->modifier);
			}
			e=e->NextSiblingElement("con");
		}
	}
}

void Connection::save() {
	TiXmlDocument xml(routeFile.c_str());
	TiXmlNode* root = xml.InsertEndChild(TiXmlElement("connections"));

	for (list<Connection*>::iterator it=connections.begin(); it!=connections.end(); it++) {
		TiXmlElement* e = root->InsertEndChild(TiXmlElement("con"))->ToElement();

		if ((*it)->from->name.empty()) e->SetAttribute("from", (*it)->from->wpID);
		else e->SetAttribute("from", (*it)->from->name.c_str());
		
		if ((*it)->to->name.empty()) e->SetAttribute("to", (*it)->to->wpID);
		else e->SetAttribute("to", (*it)->to->name.c_str());

		if ( fabs( (*it)->distance - (*it)->length() ) > .0001f )
			e->SetDoubleAttribute("dist", (*it)->distance);
		if ((*it)->modifier != 1.0f) e->SetDoubleAttribute("modifier", (*it)->modifier);
		
		if ((*it)->type==SEA) e->SetAttribute("type", "sea");
		else if ((*it)->type==RIVER) e->SetAttribute("type", "river");
		else e->SetAttribute("type", "land");

		if ((*it)->terrainType!=0) e->SetAttribute("terrainTypes", (*it)->terrainType);

		for (unsigned i=1; i<(*it)->via.size()-1; i++) {
			TiXmlElement* v = e->InsertEndChild(TiXmlElement("via"))->ToElement();
			Vector3d pos = (*it)->via[i];
			v->SetDoubleAttribute("x", pos.x);
			v->SetDoubleAttribute("y", pos.y);
		}
	}
	xml.SaveFile();
}

void Connection::draw(bool route) const {
	if (via.size()>1) {
		glBegin(GL_LINE_STRIP);
			if (!route) {
				if (type==SEA) glColor4f(0,0,1,.75f);
				if (type==LAND) glColor4f(0,1,0,.75f);
				if (type==RIVER) glColor4f(.5,.75,1,.75f);
			}

			if (via.size()==2) {
				glVertex3dv(via[0]);
				glVertex3dv(via[1]);
			}
			else {
				int last=(int)via.size()-1;
				renderCRSpline(via[0], via[0], via[1], via[2]);
				for (int i=0; i<last-2; i++)
					renderCRSpline(via[i], via[i+1], via[i+2], via[i+3]);
				renderCRSpline(via[last-2], via[last-1], via[last], via[last]);
			}
		glEnd();
	}
}
