#include "locations.h"
#include "connections.h"
#include "maptool.h"
#include <tinyxml.h>

using std::multimap;
using std::map;
using std::list;

map<string, Location*> Location::locations;
list<Location*> Location::waypoints;
string Location::locationFile;
int Location::numWaypoints=0;

void Location::create(const string& name, const Vector3d& mp) {
	Location* loc = 0;
	if (!name.empty()) {
		if (!Location::get(name)) loc = Location::add(name, mp.x, mp.y);
	}
	else loc = Location::addWP(mp.x, mp.y);
	selectedLoc = loc;

	if (loc) {
		list<Connection*> addTo;
		list<Connection*>::iterator it = Connection::connections.begin();
		Vector3d closest = (*it)->getClosestPt(mp);
		double dist = closest.distanceSq(mp);

		for (; it != Connection::connections.end(); it++) {
			Vector3d v = (*it)->getClosestPt(mp);
			double d = v.distanceSq(mp);
			if (d < dist) {
				dist = d;
				closest=v;
			}
			if (d*zoom < 50.0f) addTo.push_back(*it);
		}

		if (!addTo.empty()) {
			loc->placeAt(closest);
			for (it=addTo.begin(); it!=addTo.end();) {
				list<Connection*>::iterator tmp = it++;
				(*tmp)->insertLoc(loc);
			}
		}
	}	
}

bool Location::rename(const string& nname) {
	if (locations.find(nname) != locations.end()) return false;

	if (name.empty()) waypoints.remove(this);
	else locations.erase(name);

	name = nname;

	if (nname.empty()) {
		wpID = numWaypoints++;
		waypoints.push_back(this);
	}
	else locations[name]=this;

	return true;
}

Location* Location::get(const string& s) {
	map<string, Location*>::iterator it = locations.find(s);
	return (it!=locations.end()) ? it->second : 0;
}

Location* Location::get(const int id) {
	list<Location*>::iterator it = waypoints.begin();
	while (it!=waypoints.end() && (*it)->wpID!=id) it++;
	
	return (it!=waypoints.end()) ? *it : 0;
}

Location* Location::add(const string& s, double x, double y) {
	Location*& l = locations[s];
	if (!l) l = locations[s] = new Location(s, x, y);
	return l;
}

Location* Location::addWP(double x, double y) {
	Location* wp = new Location("", x,y);
	wp->wpID = numWaypoints++;
	waypoints.push_back (wp);	
	return wp;
}

Location* Location::pick(const Vector3d& pos) {
	Location* nearest=0;
	double closest=0;

	map<string, Location*>::iterator it = Location::locations.begin();
	for (; it!=Location::locations.end(); it++) {
		double dist = pos.distanceSq( it->second->pos );
		if (!nearest || dist<closest) {
			nearest=it->second;
			closest=dist;
		}
	}

	list<Location*>::iterator wit = Location::waypoints.begin();
	for (; wit!=Location::waypoints.end(); wit++) {
		double dist = pos.distanceSq( (*wit)->pos );
		if (!nearest || dist<closest) {
			nearest=*wit;
			closest=dist;
		}
	}

	return (closest<150.0f/zoom) ? nearest : 0;
}

void Location::getPointsInRect(const Vector3d& min, const Vector3d& max, list<Vector3d*>& l) {
	map<string, Location*>::iterator it = Location::locations.begin();
	for (; it!=Location::locations.end(); it++) {
		if (it->second->isInRect(min,max)) {
			l.push_back(&it->second->pos);

			multimap<Location*, Connection*>::iterator cit = it->second->connections.begin();
			for (; cit != it->second->connections.end(); cit++) {
				if (!cit->first->isInRect(min,max) || it->first<cit->first->name)
					cit->second->getCPointsInRect(min, max, l);
			}
		}
	}

	list<Location*>::iterator wit = Location::waypoints.begin();
	for (; wit!=Location::waypoints.end(); wit++) {
		if ((*wit)->isInRect(min,max)) {
			l.push_back(&(*wit)->pos);

			multimap<Location*, Connection*>::iterator cit = (*wit)->connections.begin();
			for (; cit != (*wit)->connections.end(); cit++) {
				if (!cit->first->isInRect(min,max) || (*wit)->name<cit->first->name ||
							((*wit)->name.empty() && cit->first->name.empty() && (*wit)->wpID<cit->first->wpID))
					cit->second->getCPointsInRect(min, max, l);
			}
		}		
	}
}

bool Location::isInRect(const Vector3d& min, const Vector3d& max) {
	return pos.x>=min.x && pos.x<=max.x && pos.y>=min.y && pos.y<=max.y;
}

void Location::placeAt(const Vector3d& p) {
	pos = p;
	multimap<Location*, Connection*>::iterator it = connections.begin();
	for (; it != connections.end(); it++) {
		if (this == it->second->from) it->second->via.front()=p;
		if (this == it->second->to) it->second->via.back()=p;
		it->second->distance = it->second->length();
	}
}

void Location::remove(Location* l) {
	multimap<Location*, Connection*>::iterator it = l->connections.begin();
	while (it != l->connections.end()) {
		multimap<Location*, Connection*>::iterator tmp = it++;
		Connection::remove(tmp->second);
	}
	if (l->name.empty()) waypoints.remove(l);
	else locations.erase(l->name);

	via.remove(l);
	delete l;
}

void Location::clear() {
	map<string, Location*>::iterator it = locations.begin();
	for (; it!=locations.end(); it++) delete it->second;
	locations.clear();

	list<Location*>::iterator wit = waypoints.begin();
	for (; wit!=waypoints.end(); wit++) delete *wit;
	waypoints.clear();

	numWaypoints=0;
}

void Location::load(const string& file) {
	TiXmlDocument xml(file.c_str());
	if (xml.LoadFile()) {
		TiXmlElement* root=xml.RootElement();

		clear();
		locationFile = file;

		TiXmlElement* e = root->FirstChildElement("loc");
		while (e) {
			Location* l = add(e->Attribute("name"));
			e->QueryDoubleAttribute("x", &l->pos.x);
			e->QueryDoubleAttribute("y", &l->pos.y);
			
			e=e->NextSiblingElement("loc");
		}

		e = root->FirstChildElement("wp");
		while (e) {
			Location* l = addWP();
			e->QueryIntAttribute("ID", &l->wpID);
			e->QueryDoubleAttribute("x", &l->pos.x);
			e->QueryDoubleAttribute("y", &l->pos.y);
			if (l->wpID >= numWaypoints) numWaypoints=l->wpID+1;
			
			e=e->NextSiblingElement("wp");
		}
	}
}

void Location::save() {
	TiXmlDocument xml(locationFile.c_str());
	TiXmlNode* root = xml.InsertEndChild(TiXmlElement("locations"));

	map<string, Location*>::iterator it=locations.begin();
	for (; it!=locations.end(); it++) {
		TiXmlElement* e = root->InsertEndChild(TiXmlElement("loc"))->ToElement();
		e->SetAttribute("name", it->second->name.c_str());
		e->SetDoubleAttribute("x", it->second->pos.x);
		e->SetDoubleAttribute("y", it->second->pos.y);
	}

	list<Location*>::iterator wit = waypoints.begin();
	for (; wit!=waypoints.end(); wit++) {
		TiXmlElement* e = root->InsertEndChild(TiXmlElement("wp"))->ToElement();
		e->SetAttribute("ID", (*wit)->wpID);
		e->SetDoubleAttribute("x", (*wit)->pos.x);
		e->SetDoubleAttribute("y", (*wit)->pos.y);
	}
	xml.SaveFile();
}
