#include "maptool.h"
#include <algorithm>
#include "platform.h"

list<SearchInfo> route;
list<Location*> via;

bool doForcedMarch=0;
bool fastestRoute=1;

void add(list<Location*>& l, Location* n) {
	list<Location*>::iterator i=l.begin();
	while (i!=l.end() && n->ctg>(*i)->ctg) i++;
	l.insert(i,n);
}

bool planSubRoute(Location* from, const Location* to, bool fastest) {
	list<Location*> open, closed;
	
	Vector3d goalPos = to->pos;

	float maxSpeed = .0f;
	vector<Transportation>::iterator tit=Transportation::transports.begin();
	for (; tit!=Transportation::transports.end(); tit++) {
		if (!tit->allowed) continue;

		float s=tit->speed;
		if (doForcedMarch && tit->type==LAND) s*=tit->fastFac;
		if (s>maxSpeed) maxSpeed=s;
	}
	float speedMod = fastest ? 1.0f/maxSpeed : 1.0f;

	Location* n = from;	
	from->prev=0;
	from->prCon=0;
	from->prTrans=0;
	n->ctg = goalPos.distance(n->pos)*speedMod;
	n->cfs=0;
	add(open, n);

	while (!open.empty()) {
		n = open.front();
		if (n==to) break;
		open.pop_front();
		closed.push_back(n);
		
		std::multimap<Location*, Connection*>::iterator it=n->connections.begin();
		for (; it!=n->connections.end(); it++) {
			if (std::find(closed.begin(), closed.end(), it->first)!=closed.end()) continue;

			float effDist = it->second->distance;

			Transportation* tran = 0;
			vector<Transportation>::iterator tt = Transportation::transports.begin();
			for (; tt!=Transportation::transports.end(); tt++) {
					if (!tt->allowed || tt->type!=it->second->type || (tt->excludedTerrain & it->second->terrainType)) continue;
					if (!tran || tran->speed<tt->speed) tran=&*tt;
			}
			if (!tran) continue;
			

			if (fastest) {
				float speed = tran->speed * it->second->modifier;
				//Upstream
				if (it->second->type==RIVER && it->first!=it->second->to)
					speed = tran->revSpeed / it->second->modifier;

				if (doForcedMarch) speed *= .5f*(1.0f+tran->fastFac);
				effDist /= speed;
			}

			list<Location*>::iterator test = std::find(open.begin(), open.end(), it->first);
			if (test!=open.end()) {
				if ((*test)->cfs < n->cfs + effDist) continue;
				else open.erase(test);
			}

			Location* potential = it->first;
			potential->cfs = n->cfs + effDist;
			potential->ctg = potential->cfs + goalPos.distance(potential->pos)*speedMod;
			potential->prev = n;
			potential->prCon = it->second;
			potential->prTrans = tran;
			add(open, potential);
		}
	}
	if (n==to) {
		if (!route.empty()) route.pop_front();
		while (n) {
			route.push_front( SearchInfo(n, n->prev, n->prCon, n->prTrans) );
			n=n->prev;
		}
		return 1;
	}
	return 0;
}

void planRoute(Location* from, Location* to, bool fastest) {
	if (!from || !to) return;

	route.clear();
	const Location* last = to;

	list<Location*>::iterator it = via.begin();
	for (; it!=via.end(); it++) {
		if (!planSubRoute(*it, last, fastest)) {
			route.clear();
			return;
		}
		last = *it;
	}
	if (!planSubRoute(from, last, fastest)) route.clear();
}

string routeToString() {
	string info;

	float segDist=0, totDist=0;
	float segDur=0, totDur=0;
	float travelled = 0;
	float forcedLeft = 1.0f;

	for (list<SearchInfo>::iterator rit=route.begin(); rit!=route.end(); rit++) {
		if (rit->prCon) {
			const Connection& con=*rit->prCon;
			segDist = con.distance;
			totDist += segDist;

			float speed = rit->prTrans->speed;
			if (con.type==RIVER && con.from==rit->location) speed = rit->prTrans->revSpeed/(con.modifier*con.modifier);
			
			speed *= con.modifier;

			if (doForcedMarch) {
				while (travelled<totDist) {
					float fspeed = speed;
					float sign=1.0f;
					float time=1.0f;
					
					if (forcedLeft>=1.0f && con.type==LAND) {
						fspeed = speed * rit->prTrans->fastFac;
						time = forcedLeft;
						sign=-1.0f;
					}
					else if (con.type==LAND) time=1.0f-forcedLeft;
					
					clamp(time, .0f, 1.0f);
					travelled += fspeed*time;
					
					if (travelled>totDist) {
						time -= (travelled-totDist)/fspeed;
						travelled=totDist;
					}
					totDur += time;
					forcedLeft+=time*sign;
					clamp(forcedLeft, .0f, 1.0f);
				}
			}
			else totDur += segDist/speed;
		}

		if ( !rit->location->name.empty() ) info+= rit->location->name;
		else {			
			if (rit->location==route.front().location) info+=LANG["start"];
			else if (rit->location==route.back().location) info+=LANG["goal"];
			else info += "[" + LANG["waypoint"] + "]";
		}
		
		info += "\n        ";
		info += rit->prTrans ? rit->prTrans->name+"\n        " : "";
		char buf[128];
		snprintf(buf, 127, "%.2f %s, %.2f %s\n", totDist, LANG["miles"].c_str(), totDur, LANG["days"].c_str());
		info += buf;
	}
	return info;
}

void centerOnRoute() {
	if (!route.empty()) {
		Vector3d min=route.front().location->pos;
		Vector3d max=min;

		for (list<SearchInfo>::iterator it=route.begin(); it!=route.end(); it++) {
			if (it->prCon) {
				for (unsigned i=0; i<it->prCon->via.size(); i++) {
					const Vector3d& p = it->prCon->via[i];
					if (p.x < min.x) min.x = p.x;
					if (p.x > max.x) max.x = p.x;
					if (p.y < min.y) min.y = p.y;
					if (p.y > max.y) max.y = p.y;
				}
			}
		}
		double wzoom = vportW/(max.x-min.x);
		double hzoom = vportH/(max.y-min.y);
		zoom = .8f * (float)(wzoom<hzoom ? wzoom : hzoom);

		Vector3d vp = Vector3d(vportW/zoom, vportH/zoom, 0);
		scrollOffset = .5*(min+max-vp);
	}
}
