#ifndef VECTOR_INCLUDED
#define VECTOR_INCLUDED

#include <cmath>
#include <cstdlib>

const double PI=3.14159265;
const double RADF=PI/180.0;

inline float deg2rad(float x) {return x*(float)RADF;}
inline double deg2rad(double x) {return x*RADF;};


template<class T>
class vector3 {
public:
	T x,y,z;

	vector3() : x(0), y(0), z(0) {};
	vector3(T s) : x(s), y(s), z(s) {}
	vector3(T* s) : x(s[0]), y(s[1]), z(s[2]) {}
	vector3(T ix, T iy, T iz) : x(ix), y(iy), z(iz) {}

	T& operator[](unsigned i) {return (&x)[i];}
	const T& operator[](unsigned i) const {return (&x)[i];}
	operator T*() {return &x;}
	operator const T*() const {return &x;}
	
	vector3<T>  operator-() const {return vector3<T>(-x, -y, -z);}
	vector3<T>  operator*(T s) const {return vector3<T>(s*x, s*y, s*z);}
	vector3<T>  operator/(T s) const {s=1/s; return vector3<T>(x*s, y*s, z*s);}
	vector3<T> operator/(const vector3<T>& b) {return vector3<T>(x/b.x, y/b.y, z/b.z);}
	vector3<T>  operator%(const vector3<T>& b) const {return vector3<T>(y*b.z - z*b.y, z*b.x - x*b.z, x*b.y - y*b.x);}
	vector3<T>  operator+(const vector3<T>& b) const {return vector3<T>(x+b.x, y+b.y, z+b.z);}	
	vector3<T>  operator-(const vector3<T>& b) const {return vector3<T>(x-b.x, y-b.y, z-b.z);}
	
	vector3<T>& operator=(T s) {x=y=z=s; return *this;}
	vector3<T>& operator*=(T s) {x*=s; y*=s; z*=s; return *this;}	
	vector3<T>& operator/=(T s) {s=1/s; x*=s; y*=s; z*=s; return *this;}
	vector3<T>& operator/=(const vector3<T>& b) {x/=b.x; y/=b.y; z/=b.z; return *this;}
	vector3<T>& operator+=(const vector3<T>& b) {x+=b.x; y+=b.y; z+=b.z; return *this;}
	vector3<T>& operator-=(const vector3<T>& b) {x-=b.x; y-=b.y; z-=b.z; return *this;}
	vector3<T>& operator%=(const vector3<T>& b) {T tmpx=x, tmpy=y; x=y*b.z-z*b.y; y=z*b.x-tmpx*b.z; z=tmpx*b.y-tmpy*b.x; return *this;}
	bool operator==(const vector3<T>& v) {return (x==v.x && y==v.y && z==v.z);}
	bool operator!=(const vector3<T>& v) {return !this->operator==(v);}
	
	T operator*(const vector3<T>& b) const {return x*b.x + y*b.y + z*b.z;}
	T distance(const vector3<T>& b) const {return (*this-b).len();}
	T distanceSq(const vector3<T>& b) const {return (*this-b).lenSq();}
	T lenSq() const {return x*x + y*y + z*z;}
	T len() const {return sqrt(lenSq());}
	vector3<T>& normalize() {return *this/=len();}

	void random(T xlim, T ylim, T zlim) {
		x= xlim*(1-(float)(rand()%2000)/1000);
		y= ylim*(1-(float)(rand()%2000)/1000);
		z= zlim*(1-(float)(rand()%2000)/1000);
	}
};

template<class T>
inline vector3<T> operator*(T s, const vector3<T>& v) {return v*s;}

typedef vector3<float> Vector3;
typedef vector3<double> Vector3d;

template<class T>
class vector4 {
public:
	T x,y,z,w;

	vector4<T>() {};
	vector4<T>(float s) : x(s), y(s), z(s), w(s) {};
	vector4<T>(float* s) : x(s[0]), y(s[1]), z(s[2]), w(s[3]) {};
	vector4<T>(vector3<T>& v, float iw=0) : x(v.x), y(v.y), z(v.z), w(iw) {}
	vector4<T>(float ix, float iy, float iz, float iw) : x(ix), y(iy), z(iz), w(iw) {};

	T& operator[](unsigned i) {return (&x)[i];}
	const T& operator[](unsigned i) const {return (&x)[i];}
	operator T*() {return &x;}
	
	vector4<T> operator-() const {return vector4<T>(-x, -y, -z, -w);}
	vector4<T> operator*(float s) const {return vector4<T>(s*x, s*y, s*z, s*w);}
	vector4<T> operator/(float s) const {s=1/s; return vector4<T>(x*s, y*s, z*s, w*s);}
	vector4<T> operator+(const vector4<T>& b) const {return vector4<T>(x+b.x, y+b.y, z+b.z, w+b.w);}	
	vector4<T> operator-(const vector4<T>& b) const {return vector4<T>(x-b.x, y-b.y, z-b.z, w-b.w);}
	
	vector4<T>& operator=(float s) {x=y=z=w=s; return *this;}
	vector4<T>& operator*=(float s) {x*=s; y*=s; z*=s; w*=s; return *this;}	
	vector4<T>& operator/=(float s) {s=1/s; x*=s; y*=s; z*=s; w*=s; return *this;}
	vector4<T>& operator+=(const vector4<T>& b) {x+=b.x; y+=b.y; z+=b.z; w+=b.w; return *this;}
	vector4<T>& operator-=(const vector4<T>& b) {x-=b.x; y-=b.y; z-=b.z; w-=b.w; return *this;}
	
	float operator*(const vector3<T>& b) const {return x*b.x + y*b.y + z*b.z;}
	float operator*(const vector4<T>& b) const {return x*b.x + y*b.y + z*b.z + w*b.w;}
	float lensq() const  {return x*x + y*y + z*z + w*w;}
	float len() const {return sqrt(lensq());}
	vector4& normalize() {return *this/=len();}

	void random(T xlim, T ylim, T zlim, T wlim) {
		x= xlim*(1-(float)(rand()%2000)/1000);
		y= ylim*(1-(float)(rand()%2000)/1000);
		z= zlim*(1-(float)(rand()%2000)/1000);
		w= wlim*(1-(float)(rand()%2000)/1000);
	}
};

typedef vector4<float> Vector4;

#endif
